import importlib
import inspect
import json
import logging
import os.path
import pathlib
import pkgutil
import shutil
import sys
from copy import deepcopy
from typing import Any, Callable

from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from hatchling.plugin import hookimpl

from beaker_kernel.lib.context import BaseContext


logger = logging.getLogger("bunsen_build")


class BuildError(Exception):
    pass


class BunsenPackageHook(BuildHookInterface):
    PLUGIN_NAME = "bunsen_package"
    _BUILD_DIR = "build"

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:

        if os.path.exists(self._BUILD_DIR):
           shutil.rmtree(self._BUILD_DIR)
        os.makedirs(self._BUILD_DIR)

        if "shared-data" not in build_data:
            build_data["shared_data"] = {}

        if self.target_name == "wheel":
            # Register Beaker subcommand
            # TODO: Move this to be a function in beaker for consistency?
            subcommand_hook = {
                "group_name": "bunsen",
                "module": "beaker_bunsen.scripts.bunsen",
                "entry_point": "cli_commands",
            }
            subcommand_hook_path = os.path.join(self._BUILD_DIR, f"subcommand_hook.json")
            with open(subcommand_hook_path, 'w') as subcommand_hook_file:
                json.dump(subcommand_hook, subcommand_hook_file, indent=2)

            build_data["shared_data"][subcommand_hook_path] = f"share/beaker/commands/beaker_bunsen.json"


@hookimpl
def hatch_register_build_hook():
    return BunsenPackageHook



import numpy as np
from dataclasses import dataclass
from numpy.typing import NDArray
from typing import Any, Union, Sequence, TypedDict, TYPE_CHECKING, TypeAlias
from typing_extensions import Self
from urllib.parse import urlparse, ParseResult


# Define types
class DefaultType:
    """Default sentinel class"""
    def __bool__(self):
        return False

    def __repr__(self) -> str:
        return 'Default'

# Default sentinel for indicating that default arguments should be used
Default: DefaultType = DefaultType()

RecordID: TypeAlias = Union[str, int]

RecordContent: TypeAlias = str

NDImage: TypeAlias = NDArray[Union[np.uint, np.int_, np.float_]]
RawImage: TypeAlias = Union[Sequence[int], Sequence[float]]
Image: TypeAlias = Union[bytearray, bytes, RawImage, NDImage]
Metadata: TypeAlias = dict
Embedding: TypeAlias = Union[Sequence[float], Sequence[int], np.ndarray]


class URI(str):
    _parts: ParseResult
    scheme: str
    netloc: str
    path: str
    params: str
    query: str
    fragment: str

    def __new__(cls, object: Any) -> Self:
        if isinstance(object, cls):
            return object
        if object is None:
            return None
        self = str.__new__(cls, object, )
        self._parts = urlparse(self)
        return self

    def __getattr__(self, name) -> Any:
        if name in ParseResult._fields:
            return getattr(self._parts, name)
        raise AttributeError(f"Attribute {name} does not exist on object {repr(self)}.")

@dataclass
class Record:
    id: RecordID
    embedding: Embedding|None = None
    content: RecordContent|None = None
    image: Image|None = None
    uri: str|URI|None = None
    metadata: Metadata|None = None

    def __post_init__(self, *args, **kwargs):
        self.uri = URI(self.uri)

RecordBundle = Sequence[Record]


class QueryResult(TypedDict):
    record: Record
    distance: float


class QueryResponse(TypedDict):
    query: str
    matches: list[QueryResult]


class ValidationError(ValueError):
    pass



import importlib
import inspect
from typing import Protocol, runtime_checkable
from typing_extensions import Self

from .resources import Resource
from .types import URI


@runtime_checkable
class EmbeddingFunction(Protocol):
    URI_SCHEME = "embedding"
    __name__: str

    def __call__(self, resource: Resource, *args, **kwargs) -> list[float]:
        ...

    @classmethod
    def get_uri(cls, value: Self | None) -> str | None:
        if value is None:
            return None
        func_name = value.__name__
        module = inspect.getmodule(value)
        spec = module.__spec__
        return f"{cls.URI_SCHEME}://{spec.name}#{func_name}"

    @classmethod
    def from_uri(cls, uri: URI | str | None) -> Self | None:
        if uri is None:
            return None
        uri = URI(uri)
        if uri.scheme != cls.URI_SCHEME:
            raise ValueError(f"Provided scheme ({uri.scheme}) does not match expected ({cls.URI_SCHEME})")
        mod_name = uri.netloc
        func_name = uri.fragment
        module = importlib.import_module(mod_name)
        func = getattr(module, func_name, None)
        if not isinstance(func, cls):
            raise ValueError(f"Function referenced by URI `{uri}` is does not satisfy the EmbeddingFunction protocol.")



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Union, Sequence, TypedDict
from typing_extensions import Self
from numpy.typing import NDArray
import logging
import numpy as np

from ..types import Record, RecordBundle, QueryResponse
from ..protocols import EmbeddingFunction
from ..resources import Resource
from ..loaders.base import BaseLoader


logger = logging.getLogger("beaker_bunsen")


class VectorStore(ABC):

    default_partition: str
    default_embedding_function: EmbeddingFunction|None
    store_settings: Any

    def __init__(
            self,
            settings: Any = None,
            default_partition: str|None = "default",
            default_embedding_function: EmbeddingFunction|None = None,
        ) -> None:
        if default_partition is not None:
            self.default_partition = default_partition
        else:
            self.default_partition = "default"
        self.default_embedding_function = default_embedding_function
        self.store_settings = settings

    @abstractmethod
    def get_partitions(self) -> list[str]:
        ...

    @abstractmethod
    def get_record(
        self,
        id: Any,
        partition: str|None = None,
    ) -> Record:
        ...

    @abstractmethod
    def get_records(
        self,
        ids: Any,
        partition: str|None = None,
    ) -> Sequence[Record]:
        ...

    @abstractmethod
    def get_all(
        self,
        partition: str|None = None,
    ) -> Sequence[Record]:
        ...

    @abstractmethod
    def add_record(
        self,
        record: Record,
        partition: str|None = None
    ):
        ...

    @abstractmethod
    def add_records(
        self,
        bundle: RecordBundle,
        partition: str|None = None
    ):
        ...

    @abstractmethod
    def update_record(
        self,
        record: Record,
        partition: str|None = None
    ):
        ...

    @abstractmethod
    def update_records(
        self,
        bundle: RecordBundle,
        partition: str|None = None
    ):
        ...

    @abstractmethod
    def query(
        self,
        query_string: str,
        partition: str | None = None,
        limit: int = -1,
        **kwargs,
        # TODO: Add extra filtering, as over metadata/subpartition/attribute/etc
    ) -> QueryResponse:
        ...

    @abstractmethod
    def query_multi(
        self,
        query_strings: list[str],
        partition: str | None = None,
        limit: int = -1,
        **kwargs,
    ) -> Sequence[QueryResponse]:
        ...

    @abstractmethod
    def clone(self, **kwargs) -> Self:
        ...

    @abstractmethod
    def save_to(
        self,
        destination: str,
        *args,
        **kwargs,
    ) -> str:
        ...




from ..protocols import EmbeddingFunction
from ..types import DefaultType, Default, ValidationError
from ..resources import Resource, find_splitter_for_resource
from ..util.logging import logger


class Embedder:
    embedding_function: EmbeddingFunction | None
    chunk_size: int = 2000
    chunk_overlap: int = 100

    def __init__(
            self,
            embedding_function: EmbeddingFunction | None = None,
            chunk_size: int | DefaultType = Default,
            chunk_overlap: int | DefaultType = Default,
        ) -> None:

        self.embedding_function = embedding_function
        self.chunk_size = chunk_size if chunk_size is not Default else self.__class__.chunk_size
        self.chunk_overlap = chunk_overlap if chunk_overlap is not Default else self.__class__.chunk_overlap

    def embed(
        self,
        resource: Resource,
        embedding_function: EmbeddingFunction = None,
        chunk_size: int | DefaultType = Default,
        chunk_overlap: int | DefaultType = Default,
    ):
        if embedding_function is None:
            embedding_function = self.embedding_function
        if chunk_size is Default:
            chunk_size = self.chunk_size
        if chunk_overlap is Default:
            chunk_overlap = self.chunk_overlap
        splitter = find_splitter_for_resource(resource, chunk_size=chunk_size, chunk_overlap=chunk_overlap)
        try:
            records = resource.as_records(splitter=splitter)
            # If there is a custom embedding function, apply that for each record
            if embedding_function:
                for record in records:
                    record.embedding = embedding_function(resource)
                    yield record
            else:
                yield from records
        except ValidationError:
            # Skip records that do not validate
            # TODO: Flags or similar for advanced control of validation error behavior? Default for now is to skip.
            pass



import importlib
import inspect
import json
import logging
import os.path
import pathlib
import pkgutil
import shutil
import sys
from copy import deepcopy
from typing import Any, Callable

import toml
from hatchling.builders.hooks.plugin.interface import BuildHookInterface
from hatchling.plugin import hookimpl

from beaker_kernel.lib.context import BaseContext
from ..corpus.corpus import Corpus
from ..corpus.types import URI
from ..corpus.loaders.code_library_loader import RCRANLocalCache
from ..corpus.loaders.schemes import RCranScheme, unmap_scheme
from ..corpus.vector_stores.chromadb_store import ZippedChromaDBStore


logger = logging.getLogger("bunsen_build")


class BuildError(Exception):
    pass


class BuildConfigError(BuildError):
    pass


class BunsenContextConfig:
    _CONFIG_KEYS_TO_SAVE = [
        "documentation_paths",
        "examples_paths",
        "libraries",
    ]
    _CONFIG_KEYS_TO_IGNORE = [
        "require-runtime-dependencies",
    ]

    _build_config: dict[str, Any]
    locations: list[URI]

    library_descriptions: dict[str, str]
    libraries: dict[str, dict[str, list[str]]] = {}

    def library_extraction(obj: dict[str, dict[str, str]]) -> list[str]:
        locations = []
        for lang, entries in obj.items():
            scheme = unmap_scheme(lang)
            if scheme is None:
                raise BuildError("Unknown language added")
            locations.extend(f"{scheme.URI_SCHEME}:{location}" for location in entries.keys())
        return locations

    CONFIG_LOCATION_MAP: dict[str, Callable[[Any], list[str]]] = {
        "documentation_path": lambda obj: [f"documentation:{obj}"],
        "documentation_paths": lambda obj: [f"documentation:{loc}" for loc in obj],
        "examples_path": lambda obj: [f"examples:{obj}"],
        "examples_paths": lambda obj: [f"examples:{loc}" for loc in obj],
        "libraries": library_extraction,
    }


    def __init__(self, build_config: dict[str, Any]) -> None:

        if "documentation_path" in build_config and "documentation_paths" in build_config:
            raise BuildConfigError(f"Both 'documentation_path' and 'documentation_paths' defined. Only one can be defined at a time.")
        if "examples_path" in build_config and "examples_paths" in build_config:
            raise BuildConfigError(f"Both 'example_path' and 'example_paths' defined. Only one can be defined at a time.")

        self._build_config = deepcopy(build_config)
        self.locations = []

        for config_opt, transformer in self.CONFIG_LOCATION_MAP.items():
            if config_opt in build_config:
                config_value = build_config.get(config_opt)
                locations = transformer(config_value)
                uris = map(URI, locations)
                self.locations.extend(uris)

        self.locations.extend(map(URI, build_config.get("locations", [])))

    @classmethod
    def from_pyproject_toml(cls, pyproject_file_path: str = "./pyproject.toml"):
        with open(pyproject_file_path) as infile:
            full_config = toml.load(infile)
        try:
            bunsen_config = full_config["tool"]["hatch"]["build"]["hooks"]["bunsen"]
        except KeyError:
            raise LookupError(f"Unable to find bunsen config in file {os.path.abspath(pyproject_file_path)}")
        return cls(bunsen_config)

    def to_json(self):
        config_dict = {
            key: value
            for key, value in self._build_config.items()
            if key in self._CONFIG_KEYS_TO_SAVE
        }
        missed_keys = set(self._build_config.keys()) - set(config_dict.keys()) - set(self._CONFIG_KEYS_TO_IGNORE)
        if missed_keys:
            missed_keys_str = "', '".join(missed_keys)
            logger.warning(f"Notice: Config option(s) '{missed_keys_str}' were defined but are not being used.")
        config_dict.update({
            "build_uris": self.locations,
        })

        # Fix up singular paths for consistency
        if "documentation_path" in config_dict:
            config_dict["documentation_paths"] = [config_dict.pop("documentation_path")]
        if "examples_path" in config_dict:
            config_dict["examples_paths"] = [config_dict.pop("examples_path")]

        return config_dict


class BunsenContextHook(BuildHookInterface):
    PLUGIN_NAME = "bunsen"
    _BUILD_DIR = "build"

    bunsen_config: BunsenContextConfig

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        self.bunsen_config = BunsenContextConfig(self.config)

        if "shared-data" not in build_data:
            build_data["shared_data"] = {}

        if self.target_name == "wheel":
            packages = self.build_config.packages

            # Build corpus, place as artifact
            # TODO: allow different contexts with different corpuses in one repo?
            corpus_path = self.build_corpus()
            for package_dir in packages:
                # Add package's parent directory to the python path so we can import the package
                sys.path.append(str(pathlib.Path(package_dir).parent))

                # TODO: check if context class exists. Only add corpus if proper context in package.
                context_slug, context_file_path = self.build_beaker_context(base_path=package_dir)
                if context_slug:
                    # Set to include the corpus
                    build_data["shared_data"][corpus_path] = f"share/beaker/corpuses/{context_slug}"

                    # Add beaker context json file(s)
                    build_data["shared_data"][context_file_path] = f"share/beaker/contexts/{context_slug}.json"

                bunsen_config_path =  os.path.join(self._BUILD_DIR, "bunsen_config.json")
                target = f"share/beaker/bunsen/{context_slug}.json"
                with open(bunsen_config_path, "w") as bunsen_config_file:
                    json.dump(self.bunsen_config.to_json(), bunsen_config_file)
                build_data["shared_data"][bunsen_config_path] = target


    def build_corpus(self) -> str:
        corpus_path = "build/corpus"
        store_path = "build/store.zip"

        if os.path.exists(corpus_path):
            shutil.rmtree(corpus_path)
        if os.path.exists(store_path):
            shutil.rmtree(store_path)

        os.makedirs(corpus_path, exist_ok=True)

        store = ZippedChromaDBStore(path=store_path)
        corpus = Corpus(store=store)

        cran_libs = [
            location.path
            for location in self.bunsen_config.locations
            if location.scheme == RCranScheme.URI_SCHEME
        ]
        with RCRANLocalCache(locations=cran_libs):
            corpus.ingest(
                self.bunsen_config.locations,
            )
            corpus.save_to_dir(corpus_path, overwrite=True)

        return corpus_path

    def build_beaker_context(self, base_path: str):

        dest_dir = "build/contexts"
        pkg_name = os.path.basename(base_path)

        slug = None
        class_name = None
        package = None

        for mod_info in pkgutil.walk_packages([base_path], prefix=f"{pkg_name}."):

            spec = mod_info.module_finder.find_spec(f"{mod_info.name}")
            mod = importlib.import_module(spec.name)

            for cls_name, cls in inspect.getmembers(mod, predicate=inspect.isclass):
                if issubclass(cls, BaseContext) and cls.__module__ == spec.name:
                    if slug:
                        raise ValueError("Can't define more than one context per package.")
                    slug = pkg_name
                    class_name = cls_name
                    package = spec.name

        if slug:
            os.makedirs(dest_dir, exist_ok=True)
            dest_file = os.path.join(dest_dir, f"{slug}.json")
            context_file_contents = {
                "slug": slug,
                "package": package,
                "class_name": class_name,
            }
            with open(dest_file, "w") as context_file:
                json.dump(context_file_contents, context_file, indent=2)
            return slug, dest_file
        else:
            return None, None


    def test_examples(self, examples):
        # TODO: Finish this
        return []



@hookimpl
def hatch_register_build_hook():
    return BunsenContextHook



from packaging.version import Version
from urllib.parse import urlparse

from beaker_bunsen.corpus.resources import isCodeResource, ResourceType
from beaker_bunsen.corpus.loaders.code_library_loader import PythonLibraryLoader, RCRANSourceLoader, RCRANLocalCache


def test_python_module_discovery():
    python_loader = PythonLibraryLoader()

    os_records = list(python_loader.discover(locations=["os"]))
    requests_records = list(python_loader.discover(locations=["requests"]))
    combined_records = list(python_loader.discover(locations=["os", "requests"]))

    requests_uris = [record.uri for record in requests_records]
    requests_schemes = [urlparse(uri).scheme for uri in requests_uris]
    requests_modules = [urlparse(uri).path for uri in requests_uris]
    top_requests_record = requests_records[0]

    assert len(os_records) == 1
    assert len(requests_records) > 5  # We don't really know how many files/submodules will always make up requests, but this limit should be relatively future proof.
    assert len(os_records) + len(requests_records) == len(combined_records)
    assert os_records[0].uri == 'py-mod:os'
    assert isinstance(top_requests_record.uri, str)
    assert isinstance(top_requests_record.id, str)
    assert top_requests_record.uri
    assert top_requests_record.id
    assert 'requests.adapters' in requests_modules
    assert all([scheme == "py-mod" for scheme in requests_schemes])
    assert all([module.startswith("requests") for module in requests_modules])


def test_python_load():
    python_loader = PythonLibraryLoader()

    os_record = list(python_loader.discover(locations=["os"]))[0]
    uri = os_record.uri
    source = python_loader.read(uri)

    assert len(source) > 0
    assert isinstance(source, str)
    assert source.startswith('r"""OS')
    assert "\nenviron = " in source
    assert "def makedirs(" in source


def test_exclusions_init():
    excluded_python_loader = PythonLibraryLoader(locations=["requests", "!requests._internal_utils", "!__version__"])
    excluded_requests_records = list(excluded_python_loader.discover())
    full_python_loader = PythonLibraryLoader(locations=["requests"])
    full_requests_records = list(full_python_loader.discover())

    excluded_record_uris = [record.uri for record in excluded_requests_records]
    full_record_uris = [record.uri for record in full_requests_records]

    assert 'py-mod:requests._internal_utils' in full_record_uris
    assert 'py-mod:requests.__version__' in full_record_uris
    assert 'py-mod:requests._internal_utils' not in excluded_record_uris
    assert 'py-mod:requests.__version__' not in excluded_record_uris

    assert set(full_record_uris) - set(excluded_record_uris) == set(['py-mod:requests._internal_utils', 'py-mod:requests.__version__'])


def test_exclusions_discover():
    excluded_python_loader = PythonLibraryLoader()
    excluded_requests_records = list(excluded_python_loader.discover(locations=["requests", "!requests._internal_utils", "!__version__"]))
    full_python_loader = PythonLibraryLoader()
    full_requests_records = list(full_python_loader.discover(locations=["requests"]))

    excluded_record_uris = [record.uri for record in excluded_requests_records]
    full_record_uris = [record.uri for record in full_requests_records]

    assert 'py-mod:requests._internal_utils' in full_record_uris
    assert 'py-mod:requests.__version__' in full_record_uris
    assert 'py-mod:requests._internal_utils' not in excluded_record_uris
    assert 'py-mod:requests.__version__' not in excluded_record_uris

    assert set(full_record_uris) - set(excluded_record_uris) == set(['py-mod:requests._internal_utils', 'py-mod:requests.__version__'])


def test_exclusions_base_exclusions():
    excluded_python_loader = PythonLibraryLoader(exclusions=["__version__"])
    excluded_requests_records = list(excluded_python_loader.discover(locations=["requests", "!requests._internal_utils"]))
    full_python_loader = PythonLibraryLoader()
    full_requests_records = list(full_python_loader.discover(locations=["requests"]))

    excluded_record_uris = [record.uri for record in excluded_requests_records]
    full_record_uris = [record.uri for record in full_requests_records]

    assert 'py-mod:requests._internal_utils' in full_record_uris
    assert 'py-mod:requests.__version__' in full_record_uris
    assert 'py-mod:requests._internal_utils' not in excluded_record_uris
    assert 'py-mod:requests.__version__' not in excluded_record_uris

    assert set(full_record_uris) - set(excluded_record_uris) == set(['py-mod:requests._internal_utils', 'py-mod:requests.__version__'])


def test_r_cran_source_loader():
    loader = RCRANSourceLoader()

    resources = list(loader.discover(
        locations=["jsonlite"]
    ))
    resource_types = set([resource.resource_type for resource in resources])

    assert len(RCRANLocalCache.remote_package_cache) > 0
    assert "jsonlite" in RCRANLocalCache.remote_package_cache
    assert Version(RCRANLocalCache.remote_package_cache["jsonlite"]["version"]) > Version("1.8")
    assert len(resources) > 0
    assert resource_types == {ResourceType.Code, ResourceType.Documentation, ResourceType.Example}



def test_r_cran_source_loader_w_filter():
    loader = RCRANSourceLoader()

    resources = list(loader.discover(
        locations=["jsonlite"],
        filter=isCodeResource,
    ))
    resource_types = set([resource.resource_type for resource in resources])

    assert len(RCRANLocalCache.remote_package_cache) > 0
    assert "jsonlite" in RCRANLocalCache.remote_package_cache
    assert Version(RCRANLocalCache.remote_package_cache["jsonlite"]["version"]) > Version("1.8")
    assert len(resources) > 0
    assert resource_types == {ResourceType.Code,}



import asyncio
import json
import logging
import os
from typing import TYPE_CHECKING, Any, Dict, List
from dataclasses import dataclass

from beaker_kernel.lib.autodiscovery import LIB_LOCATIONS
from beaker_kernel.lib.context import BaseContext
from beaker_kernel.lib.subkernels.base import BaseSubkernel
from beaker_kernel.lib.subkernels import autodiscover_subkernels

from .bunsen_agent import BunsenAgent
from .corpus.corpus import Corpus
from .corpus.types import QueryResult, QueryResponse

if TYPE_CHECKING:
    from beaker_kernel.kernel import LLMKernel


logger = logging.getLogger(__name__)


class EnvironmentSerializer:
    pass


@dataclass
class BunsenLibrary:
    name: str
    description: str


@dataclass
class BunsenLanguageLibraries:
    libraries: dict[str, BunsenLibrary]
    subkernel_cls: type[BaseSubkernel] | None


class BunsenContext(BaseContext):
    agent_cls: BunsenAgent
    corpus: Corpus

    enabled_subkernels: list[str]
    bunsen_config: dict[str, Any] | None
    environment_serializer: EnvironmentSerializer
    _subkernel_state: dict[str, Any] | None

    enabled_subkernels = ["python3"]

    PROMPT_INTRO: str = """
You are a diligent and thorough coding assistant that consistently delivers accurate and reliable responses to user
instructions.
""".strip()
    PROMPT_OUTRO: str = """
Please answer any user queries or perform user instructions to the best of your ability.
You should look up any information you need using the tools provided. If there is a chance that a tool has more
up-to-date knowledge then what you have been trained on, you should always use the tool to ensure you have as much
information as possible. This applies if you are are answering a question, writing code, or retrieving data to pass to a
tool.
Do not guess if you are not sure of an answer. If you are unsure of an answer, always be clear of which information you
unsure of. The user is trusting you to give factual answers and to not just guess.
""".strip()

    @classmethod
    def default_payload(cls) -> str:
        return "{}"

    def __init__(
        self,
        beaker_kernel: "LLMKernel",
        config: Dict[str, Any],
    ) -> None:
        self._subkernel_state = None
        super().__init__(beaker_kernel, self.agent_cls, config)
        corpus_dir = self.corpus_location()
        self.corpus = Corpus.from_dir(corpus_dir)
        self.bunsen_config = self._load_bunsen_config()

    def _load_bunsen_config(self) -> dict[str, Any]:
        for location in LIB_LOCATIONS:
            target = os.path.join(location, "bunsen", f"{self.slug}.json")
            if os.path.exists(target):
                with open(target) as bunsen_config_file:
                    bunsen_config = json.load(bunsen_config_file)
                return bunsen_config
        return {}

    async def setup(self, context_info=None, parent_header=None):
        return await super().setup(context_info, parent_header)

    async def post_execute(self, message):
        try:
            self._subkernel_state = await self.get_subkernel_state()
        except Exception as err:
            logger.error(msg=str(err), exc_info=err)

    def corpus_location(self) -> str | None:
        for location in LIB_LOCATIONS:
            corpus_path = os.path.join(location, "corpuses", self.slug)
            if os.path.isdir(corpus_path):
                return corpus_path
        return None


    @classmethod
    def available_subkernels(cls) -> List[BaseSubkernel]:
        if cls.enabled_subkernels:
            return cls.enabled_subkernels
        else:
            return super().available_subkernels()

    async def subkernel_state(self) -> dict[str, Any]:
        if self._subkernel_state:
            return self._subkernel_state
        else:
            result = await self.get_subkernel_state()
            return result

    @property
    def libraries(self) -> dict[str, BunsenLanguageLibraries]:
        """
        Enrich the libraries with information about the subkernel, if available.
        """
        libraries: dict[str, dict[str, str]] = self.bunsen_config.get("libraries", {})
        installed_subkernels = autodiscover_subkernels()
        output = {}
        for lang, libs in libraries.items():
            lib = BunsenLanguageLibraries(
                libraries={name: BunsenLibrary(name, description) for name, description in libs.items()},
                subkernel_cls=installed_subkernels.get(lang, None),
            )
            output[lang] = lib
        return output

    @property
    def library_description(self) -> str:
        """
        """
        if not self.libraries:
            return ""

        output = [
            "Consider the following installed libraries as your primary toolset for accomplishing tasks.",
            "You should always default to using these libraries if you can, but may use tools from other libraries if needed.",
            "",
        ]
        for lang in self.libraries.values():
            # Only include libraries that match the current subkernel
            if lang.subkernel_cls and isinstance(self.subkernel, lang.subkernel_cls):
                output.append(f'''{lang.subkernel_cls.DISPLAY_NAME} (Jupyter kernel "{lang.subkernel_cls.KERNEL_NAME}") libraries:''')
                for lib in lang.libraries.values():
                    output.append( f'''  {lib.name}:''')
                    output.extend([f'''    {line.strip()}''' for line in lib.description.splitlines() if line.strip()])
                    output.append('')  # Newline between libraries
                output.append('')  # Newline between subkernels

        return "\n".join(output)

    async def get_context_prompt(self):
        """
        Class that can be overridden by subclassing to provide context specific prompt information in to the autoprompt
        """
        return None

    async def get_subkerkel_state_description(self) -> str:
        state = await self.subkernel_state()
        output = []
        if state["variables"]:
            output.append("The coding environment's state currently has the following local variables:")
            output.extend(
                f"    `{var_name}`: `{var_value}`" for var_name, var_value in state["variables"].items()
            )
        if state["functions"]:
            output.append("The coding environment's state currently has the following local functions defined:")
            output.extend(
                f"    `{func_name}`: `{func_value}`" for func_name, func_value in state["functions"].items()
            )
        if state["modules"]:
            output.append("The coding environment's state currently has the following modules imported:")
            output.extend(
                f"    `{mod_name}`: `{mod_value}`" for mod_name, mod_value in state["modules"].items()
            )
        return "\n".join(output)

    async def get_documentation_string(self) -> str:
        docs = await self.get_documentation(query=self.current_llm_query)
        if docs:
            document_str = "\n\n".join(
                """
======== documentation excerpt {num}: {document_id} start ========
{document}
======== documentation excerpt {num}: {document_id} end   ========
                """.strip().format(document_id=document["record"].id, document=document["record"].content, num=num)
                for num, document in enumerate(docs, start=1)
            )
            return "\n".join([
                """Below are some excerpts from the documentation that help you. Please use them if they are relevent.""",
                document_str,
            ])
        else:
            return None


    async def get_examples(
        self,
        query=None,
        count=5,
    ) -> list[QueryResult]:
        if query is None:
            query = self.current_llm_query
        if query:
            examples = await self.query_corpus(
                query_str=query,
                partition="examples",
                limit=count,
            )
            return examples["matches"]
        return None

    async def get_documentation(
        self,
        query=None,
        count=3,
    ) -> list[QueryResult] | None:
        if query is None:
            query = self.current_llm_query

        if query:
            docs = await self.query_corpus(
                query_str=query,
                partition="documentation",
                limit=count,
            )
            return docs["matches"]
        return None

    async def get_source_code(
        self,
        query=None,
        count=3,
    ):
        matches = self.corpus.store.query(
            query_string=query,
            partition="code",
            limit=count,
        )["matches"]
        return matches

    async def build_prompt(
            self,
        ) -> str:

        state_description_future = self.get_subkerkel_state_description()
        context_prompt_future = self.get_context_prompt()
        docs_future = self.get_documentation_string()

        state_description, context_prompt, docs_result = await asyncio.gather(
            state_description_future,
            context_prompt_future,
            docs_future
        )

        prompt = [
            self.PROMPT_INTRO,
            self.library_description,
            state_description,
            docs_result,
            context_prompt,
            self.PROMPT_OUTRO,
        ]

        result = "\n".join(part for part in prompt if part)
        return result


    async def auto_context(self):
        prompt = await self.build_prompt()
        self.beaker_kernel.log("auto_context", prompt)
        return prompt


    async def query_corpus(
        self,
        query_str: str,
        partition: str = "default",
        limit: int = 5,
    ) -> QueryResponse:
        results = self.corpus.store.query(
            query_string=query_str,
            partition=partition,
            limit=limit,
        )
        return results
